import os
import base64
import requests as req
from .afas_filter import AfasFilter, afas_filters_to_query

class AfasConnector:
    """
    A class representing an AFAS Connector.

    This class provides methods to interact with AFAS Connectors through HTTP requests.
    """

    def __init__(self, connector_url: str | None = None, token: str | None = None):
        """
        Initializes an instance of the AfasConnector class.

        The AFAS connector URL and token are retrieved from environment variables.
        """
        # AFAS connector URL
        self.AFAS_CONNECTORS_URL = os.getenv('AFAS_CONNECTORS_URL') if connector_url is None else connector_url

        # Get AFAS token
        self.afas_token = base64.b64encode((os.getenv('AFAS_TOKEN') if token is None else token).encode()).decode()

        # Create AFAS headers
        self.headers = {
            'Authorization': 'AfasToken ' + self.afas_token
        }

    def get(self, endpoint: str, take=-1, skip=-1, filters: list[AfasFilter] | None=None, order_by: str | None=None, ascending=True):
        """
        Sends a GET request to the AFAS Connector.

        Args:
            endpoint (str): The endpoint of the AFAS Connector.
            take (int, optional): The number of records to retrieve. Defaults to 1000000.
            skip (int, optional): The number of records to skip. Defaults to 0.
            filters (list[AfasFilter], optional): A list of AfasFilter objects to apply as filters. Defaults to None.
            order_by (str, optional): The field to order the records by. Defaults to None.
            ascending (bool, optional): Whether to sort the records in ascending order. Defaults to True.

        Returns:
            tuple: A tuple containing the status code and the JSON response from the AFAS Connector.
        """
        # Check if endpoint starts with "/"
        if endpoint[0] != "/":
            endpoint = "/" + endpoint

        # Create the URL
        url = self.AFAS_CONNECTORS_URL + endpoint + f'?take={take}&skip={skip}'

        # Add ordering
        if order_by is not None:
            if ascending:
                url += f'&orderbyfieldids={order_by}'
            else:
                url += f'&orderbyfieldids=-{order_by}'

        # Add filters
        if filters is not None:
            url += afas_filters_to_query(filters)

        # Make the GET request
        print(f"GET {url}")
        response = req.get(url, headers=self.headers)

        if response.status_code < 200 or response.status_code >= 300:
            print(f"Request error: {response.status_code} {response.text}")
            return response.status_code, None
        
        return response.status_code, response.json()


    def put(self, endpoint: str, payload: dict):
        """
        Sends a PUT request to the AFAS Connector.

        Args:
            endpoint (str): The endpoint of the AFAS Connector.
            payload (dict): The payload to send in the request.

        Returns:
            tuple: A tuple containing the status code and the JSON response from the AFAS Connector.
        """
        # Check if endpoint starts with "/"
        if endpoint[0] != "/":
            endpoint = "/" + endpoint

        # Create the URL
        url = self.AFAS_CONNECTORS_URL + endpoint

        # Make the PUT request
        print(f"PUT {url}")
        response = req.put(url, headers=self.headers, json=payload)

        if response.status_code < 200 or response.status_code >= 300:
            print(f"Request error: {response.status_code} - {response.text}")
            return response.status_code, None
        
        # Try to parse the response as JSON
        try:
            data = response.json()
        except:
            data = response.text

        return response.status_code, data
    
    def post(self, endpoint: str, payload: dict):
        """
        Sends a POST request to the AFAS Connector.

        Args:
            endpoint (str): The endpoint of the AFAS Connector.
            payload (dict): The payload to send in the request.

        Returns:
            tuple: A tuple containing the status code and the JSON response from the AFAS Connector.
        """
        # Check if endpoint starts with "/"
        if endpoint[0] != "/":
            endpoint = "/" + endpoint

        # Create the URL
        url = self.AFAS_CONNECTORS_URL + endpoint

        # Make the POST request
        print(f"POST {url}")
        response = req.post(url, headers=self.headers, json=payload)

        if response.status_code < 200 or response.status_code >= 300:
            print(f"Request error: {response.status_code} - {response.text}")
            return response.status_code, None
        
        # Try to parse the response as JSON
        try:
            data = response.json()
        except:
            data = response.text

        return response.status_code, data




from .afas_connector import AfasConnector


class AfasObject:
    """AfasObject is a class that represents an object in AFAS Profit.

    Attributes:
        type (str): The unique identifier of the object.
        fields (dict): The fields of the object.
    """

    def __init__(self, type, fields: dict = {}):
        """Initializes AfasObject with the given parameters.

        Args:
            type (str): The unique identifier of the object.
            fields (dict, optional): The fields of the object. Defaults to {}.
        """
        self.type = type
        self.set(fields)

        self.afas_connector = AfasConnector()

    def set(self, fields: dict):
        """Sets the fields of the object.

        Args:
            fields (dict): The fields of the object.
        """
        self.fields = fields

    def payload(self):
        """Returns a dictionary representation of the object.

        Returns:
            dict: A dictionary representation of the object.
        """
        return {
            self.type: {
                "Element": {
                    "Fields": self.fields
                }
            }
        }
    
    def update(self):
        """Updates the object in AFAS Profit.

        Returns:
            tuple: A tuple containing the status code and the JSON response from the AFAS Connector.
        """
        return self.afas_connector.put(self.type, self.payload())

    def __str__(self):
        """Returns a string representation of the object.

        Returns:
            str: A string representation of the object.
        """
        return f"AfasObject(id={self.id}, name={self.name}, type={self.type}, description={self.description}, fields={self.fields})"


from setuptools import setup, find_packages

setup(
    name="afas_connector",
    version="0.1.3",
    packages=find_packages(),
    install_requires=[
        'requests >= 2.31.0',
        'python-dotenv >= 1.0.1'
    ],
    author="Robin Bakker",
    author_email="robin@bakeable.nl",
    description="AfasConnector is a Python package that simplifies interacting with AFAS Connectors via HTTP requests. It provides classes for making GET, POST, and PUT requests to AFAS Connectors and handling their responses.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/bakeable/afas-connector",
    license="MIT",
    classifiers=[
        # Trove classifiers (https://pypi.org/classifiers/)
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.12',
)



class AfasFilter:
    EQUAL_TO = 1
    GREATER_THAN_OR_EQUAL_TO = 2
    LESS_THAN_OR_EQUAL_TO = 3
    GREATER_THAN = 4
    LESS_THAN = 5
    CONTAINS_TEXT = 6
    NOT_EQUAL_TO = 7
    IS_EMPTY = 8
    IS_NOT_EMPTY = 9
    STARTS_WITH = 10
    DOES_NOT_CONTAIN_TEXT = 11
    DOES_NOT_START_WITH = 12
    ENDS_WITH = 13
    DOES_NOT_END_WITH = 14
    QUICK_FILTER = 15

    def __init__(self, field: str, value: any, operator: int = 1):
        """
        Initializes an instance of the AfasFilter class.

        Args:
            field (str): The field to filter on.
            value (any): The value(s) to filter on. Can be a single value or a list of values.
            operator (int, optional): The operator to use for the filter. Defaults to EQUAL_TO.

        Raises:
            ValueError: If the operator is invalid.
            ValueError: If the value is a list but the operator is not a list.
            ValueError: If the length of the value and operator lists are not equal.
        """

        # If value is a list, operator should be a list too and of equal length
        if isinstance(value, list):
            if not isinstance(operator, list):
                raise ValueError("Operator should be a list if value is a list")
            if len(value) != len(operator):
                raise ValueError("Value and operator should be of equal length")

        # Check if operators are valid
        if isinstance(operator, list):
            for op in operator:
                if not self.operator_is_valid(op):
                    raise ValueError("Invalid operator")
        else:
            if not self.operator_is_valid(operator):
                raise ValueError("Invalid operator")
        
        
        # Set the filter properties
        self.field = field
        self.value = value
        self.operator = operator

    def operator_is_valid(self, operator):
        if operator not in [
            self.EQUAL_TO,
            self.GREATER_THAN_OR_EQUAL_TO,
            self.LESS_THAN_OR_EQUAL_TO,
            self.GREATER_THAN,
            self.LESS_THAN,
            self.CONTAINS_TEXT,
            self.NOT_EQUAL_TO,
            self.IS_EMPTY,
            self.IS_NOT_EMPTY,
            self.STARTS_WITH,
            self.DOES_NOT_CONTAIN_TEXT,
            self.DOES_NOT_START_WITH,
            self.ENDS_WITH,
            self.DOES_NOT_END_WITH,
            self.QUICK_FILTER
        ]:
            return False
        
        return True

def afas_filters_to_query(filters: list[AfasFilter]):
    """
    Converts a list of AfasFilter objects to a query string.

    Args:
        filters (list[AfasFilter]): A list of AfasFilter objects.

    Returns:
        str: The query string.
    """
    OR_SEPARATOR = "%3B"
    AND_SEPARATOR = "%2C"

    # Initialize variables
    filterfieldids = ""
    filtervalues = ""
    operatortypes = ""

    # Add filters to the query
    for filter in filters:

        if isinstance(filter.value, list):
            # Add OR Statements to the query
            for i in range(len(filter.value)):
                filterfieldids += f'{filter.field}{OR_SEPARATOR}'
                filtervalues += f'{filter.value[i]}{OR_SEPARATOR}'
                operatortypes += f'{filter.operator[i]}{OR_SEPARATOR}'

            # Remove final %3B
            filterfieldids = filterfieldids[:-len(OR_SEPARATOR)]
            filtervalues = filtervalues[:-len(OR_SEPARATOR)]
            operatortypes = operatortypes[:-len(OR_SEPARATOR)]
        else:
            filterfieldids += f'{filter.field}'
            filtervalues += f'{filter.value}'
            operatortypes += f'{filter.operator}'

        # Add separator
        filterfieldids += AND_SEPARATOR
        filtervalues += AND_SEPARATOR
        operatortypes += AND_SEPARATOR

    # Remove final separator
    filterfieldids = filterfieldids[:-len(AND_SEPARATOR)]
    filtervalues = filtervalues[:-len(AND_SEPARATOR)]
    operatortypes = operatortypes[:-len(AND_SEPARATOR)]
    
    # Add filters to the URL
    query = f'&filterfieldids={filterfieldids}&filtervalues={filtervalues}&operatortypes={operatortypes}'

    return query